/*  -*- C++ -*- */

%skeleton "lalr1.cc"
%define "parser_class_name" "ValueExpressionParser"

%defines
%{
#include <volFields.H>
#include <surfaceFields.H>
#include <fvcGrad.H>
#include <fvcCurl.H>
#include <fvcMagSqrGradGrad.H>
#include <fvcSnGrad.H>
#include <fvcDiv.H>
#include <fvcSurfaceIntegrate.H>
#include <fvcReconstruct.H>
#include <fvcAverage.H>
#include <surfaceInterpolate.H>
#include <fvcLaplacian.H>

#include <fvMatrix.H>

#include <functional>
#include <cmath>

#include "ValueExpressionDriver.H"
#include "ValueExpressionDriverLogicalTemplates.H"
%}

%name-prefix="ve"

%parse-param { ValueExpressionDriver& driver }
%lex-param { ValueExpressionDriver& driver }

%locations
%initial-action
{
	     // Initialize the initial location.
	     //     @$.begin.filename = @$.end.filename = &driver.file;
};

%debug
%error-verbose

%union
{
    Foam::scalar val;
    Foam::vector *vec;
    Foam::string *name;
    Foam::string *vname;
    Foam::string *fsname;
    Foam::string *fvname;
    Foam::string *setname;
    Foam::volVectorField *vfield;
    Foam::volScalarField *sfield;
    Foam::surfaceScalarField *fsfield;
    Foam::surfaceVectorField *fvfield;
};

%token <name>   TOKEN_SID   "scalarID"
%token <vname>  TOKEN_VID   "vectorID"
%token <fsname> TOKEN_FSID  "faceScalarID"
%token <fvname> TOKEN_FVID  "faceVectorID"
%token <setname> TOKEN_SETID "cellSetID" 
%token <val>    TOKEN_NUM   "number"
%token <vec>    TOKEN_VEC   "vector"
%type  <sfield>    exp        "expression"
%type  <sfield>    lexp       "lexpression"
%type  <vfield>    vexp       "vexpression"
%type  <fsfield>   fsexp      "fsexpression"
%type  <fvfield>   fvexp      "fvexpression"
%type  <vfield>    vector     

%token TOKEN_VECTOR

%token TOKEN_TRUE
%token TOKEN_FALSE

%token TOKEN_pi
%token TOKEN_rand
%token TOKEN_randNormal
%token TOKEN_position
%token TOKEN_fposition
%token TOKEN_face
%token TOKEN_area
%token TOKEN_volume
%token TOKEN_dist
%token TOKEN_set

%token TOKEN_div
%token TOKEN_grad
%token TOKEN_curl
%token TOKEN_snGrad
%token TOKEN_magSqrGradGrad;
%token TOKEN_laplacian

%token TOKEN_integrate
%token TOKEN_sum
%token TOKEN_interpolate
%token TOKEN_average
%token TOKEN_reconstruct

%token TOKEN_pow
%token TOKEN_log
%token TOKEN_exp
%token TOKEN_mag
%token TOKEN_sin
%token TOKEN_cos
%token TOKEN_tan
%token TOKEN_min
%token TOKEN_max
%token TOKEN_sqr
%token TOKEN_sqrt

%left '?' ':'
%left TOKEN_OR
%left TOKEN_AND
%left TOKEN_EQ TOKEN_NEQ
%left TOKEN_LEQ TOKEN_GEQ '<' '>'
%left '-' '+'
%left '*' '/' '&' '^'
%left TOKEN_NEG TOKEN_NOT
// %right '^'
%left '.'

%printer             { debug_stream () << *$$; } "scalarID" "vectorID" "faceScalarID" "faceVectorID" "cellSetID"
%printer             { Foam::OStringStream buff; buff << *$$; debug_stream () << buff.str().c_str(); } "vector"
%destructor          { delete $$; } "scalarID" "faceScalarID" "faceVectorID" "vectorID" "vector" "expression" "vexpression" "fsexpression" "fvexpression" "lexpression"  "cellSetID"
%printer             { debug_stream () << $$; } "number" 
%printer             { debug_stream () << $$->name().c_str(); } "expression"  "vexpression" "lexpression" "fsexpression" "fvexpression"


%%
%start unit;

unit:   exp                     { driver.setScalarResult($1);  }
        | vexp                  { driver.setVectorResult($1);  }
        | lexp                  { driver.setLogicalResult($1); }
;

vexp:   vector                  { $$ = $1; }
        | vexp '+' vexp 		{ $$ = new Foam::volVectorField(*$1 + *$3); delete $1; delete $3; }
        | exp '*' vexp 		        { $$ = new Foam::volVectorField(*$1 * *$3); delete $1; delete $3; }
        | vexp '*' exp 		        { $$ = new Foam::volVectorField(*$1 * *$3); delete $1; delete $3; }
        | vexp '/' exp 		        { $$ = new Foam::volVectorField(*$1 / *$3); delete $1; delete $3; }
        | vexp '^' vexp 		{ $$ = new Foam::volVectorField(*$1 ^ *$3); delete $1; delete $3; }
        | vexp '-' vexp 		{ $$ = new Foam::volVectorField(*$1 - *$3); delete $1; delete $3;}
        | '-' vexp %prec TOKEN_NEG 	        { $$ = new Foam::volVectorField(-*$2); delete $2; }
        | '(' vexp ')'		        { $$ = $2; }  
        | lexp '?' vexp ':' vexp        { $$ = driver.doConditional($1,$3,$5,driver.makeVectorField(Foam::vector())); delete $1; delete $3; delete $5; }
        | TOKEN_position '(' ')'        { $$ = driver.makePositionField(); }
        | TOKEN_laplacian '(' fsexp ',' vexp ')'      { $$ = new Foam::volVectorField(Foam::fvc::laplacian(*$3,*$5)); delete $3; delete $5; }
        | TOKEN_average '(' fvexp ')'        { $$ = new Foam::volVectorField(Foam::fvc::average(*$3)); delete $3; }
        | TOKEN_integrate '(' fvexp ')'        { $$ = new Foam::volVectorField(Foam::fvc::surfaceIntegrate(*$3)); delete $3; }
        | TOKEN_sum '(' fvexp ')'        { $$ = new Foam::volVectorField(Foam::fvc::surfaceSum(*$3)); delete $3; }
        | TOKEN_grad '(' exp ')'        { $$ = new Foam::volVectorField(Foam::fvc::grad(*$3)); delete $3; }
        | TOKEN_reconstruct '(' fsexp ')'        { $$ = new Foam::volVectorField(Foam::fvc::reconstruct(*$3)); delete $3; }
        | TOKEN_curl '(' vexp ')'      { $$ = new Foam::volVectorField(Foam::fvc::curl(*$3)); delete $3; }
        | TOKEN_div '(' fsexp ',' vexp ')'      { $$ = new Foam::volVectorField(Foam::fvc::div(*$3,*$5)); delete $3; delete $5; }
        | TOKEN_VID {
            $$=driver.getVectorField(*$1);
                    }
;

fsexp:    fsexp '+' fsexp 		{ $$ = new Foam::surfaceScalarField(*$1 + *$3); delete $1; delete $3; }
        | fsexp '*' fsexp 		{ $$ = new Foam::surfaceScalarField(*$1 * *$3); delete $1; delete $3; }
        | fvexp '&' fvexp 		{ $$ = new Foam::surfaceScalarField(*$1 & *$3); delete $1; delete $3; }
        | fsexp '/' fsexp 		{ $$ = new Foam::surfaceScalarField(*$1 / *$3); delete $1; delete $3; }
        | fsexp '-' fsexp 		{ $$ = new Foam::surfaceScalarField(*$1 - *$3); delete $1; delete $3;}
        | '-' fsexp %prec TOKEN_NEG 	        { $$ = new Foam::surfaceScalarField(-*$2); delete $2; }
        | '(' fsexp ')'		        { $$ = $2; }  
        | TOKEN_mag '(' fsexp ')'       { $$ = new Foam::surfaceScalarField(Foam::mag(*$3)); delete $3; }
        | TOKEN_mag '(' fvexp ')'       { $$ = new Foam::surfaceScalarField(Foam::mag(*$3)); delete $3; }
        | TOKEN_area '(' ')'            { $$ = driver.makeAreaField(); }
        | TOKEN_snGrad '(' exp ')'        { $$ = new Foam::surfaceScalarField(Foam::fvc::snGrad(*$3)); delete $3; }
        | TOKEN_interpolate '(' exp ')'        { $$ = new Foam::surfaceScalarField(Foam::fvc::interpolate(*$3)); delete $3; }
        | TOKEN_FSID {
            $$=driver.getFaceScalarField(*$1);
                    }
;
 
fvexp:    fvexp '+' fvexp 		{ $$ = new Foam::surfaceVectorField(*$1 + *$3); delete $1; delete $3; }
        | fsexp '*' fvexp 		{ $$ = new Foam::surfaceVectorField(*$1 * *$3); delete $1; delete $3; }
        | fvexp '*' fsexp 		{ $$ = new Foam::surfaceVectorField(*$1 * *$3); delete $1; delete $3; }
        | fvexp '^' fvexp 		{ $$ = new Foam::surfaceVectorField(*$1 ^ *$3); delete $1; delete $3; }
        | fvexp '/' fsexp 		{ $$ = new Foam::surfaceVectorField(*$1 / *$3); delete $1; delete $3; }
        | fvexp '-' fvexp 		{ $$ = new Foam::surfaceVectorField(*$1 - *$3); delete $1; delete $3;}
        | '-' fvexp %prec TOKEN_NEG 	        { $$ = new Foam::surfaceVectorField(-*$2); delete $2; }
        | '(' fvexp ')'		        { $$ = $2; }  
        | TOKEN_fposition '(' ')'       { $$ = driver.makeFacePositionField(); }
        | TOKEN_face '(' ')'            { $$ = driver.makeFaceField(); }
        | TOKEN_snGrad '(' vexp ')'        { $$ = new Foam::surfaceVectorField(Foam::fvc::snGrad(*$3)); delete $3; }
        | TOKEN_interpolate '(' vexp ')'        { $$ = new Foam::surfaceVectorField(Foam::fvc::interpolate(*$3)); delete $3; }
        | TOKEN_FVID {
            $$=driver.getFaceVectorField(*$1);
                    }
;
 
exp:	TOKEN_NUM		{ $$ = driver.makeScalarField($1); }
        | exp '+' exp 		{ $$ = new Foam::volScalarField(*$1 + *$3); delete $1; delete $3; }
        | exp '-' exp 		{ $$ = new Foam::volScalarField(*$1 - *$3); delete $1; delete $3; }
        | exp '*' exp 		{ $$ = new Foam::volScalarField(*$1 * *$3); delete $1; delete $3; }
        | exp '/' exp 		{ $$ = new Foam::volScalarField(*$1 / *$3); delete $1; delete $3; }
        | TOKEN_pow '(' exp ',' TOKEN_NUM ')'		{ $$ = new Foam::volScalarField(Foam::pow(*$3, $5)); delete $3; }
        | TOKEN_log '(' exp ')'       { $$ = new Foam::volScalarField(Foam::log(*$3)); delete $3; }
        | TOKEN_exp '(' exp ')'       { $$ = new Foam::volScalarField(Foam::exp(*$3)); delete $3; }
        | vexp '&' vexp 	{ $$ = new Foam::volScalarField(*$1 & *$3); delete $1; delete $3;}
        | '-' exp %prec TOKEN_NEG 	{ $$ = new Foam::volScalarField(-*$2); delete $2; }
	| '(' exp ')'		{ $$ = $2; }
        | TOKEN_sqr '(' exp ')'       { $$ = new Foam::volScalarField(Foam::sqr(*$3)); delete $3; }
        | TOKEN_sqrt '(' exp ')'       { $$ = new Foam::volScalarField(Foam::sqrt(*$3)); delete $3; }
        | TOKEN_sin '(' exp ')'       { $$ = new Foam::volScalarField(Foam::sin(*$3)); delete $3; }
        | TOKEN_cos '(' exp ')'       { $$ = new Foam::volScalarField(Foam::cos(*$3)); delete $3; }
        | TOKEN_tan '(' exp ')'       { $$ = new Foam::volScalarField(Foam::tan(*$3)); delete $3; }
        | TOKEN_min '(' exp ')'       { $$ = driver.makeScalarField(Foam::min(*$3).value()); delete $3; }
        | TOKEN_max '(' exp ')'       { $$ = driver.makeScalarField(Foam::max(*$3).value()); delete $3; }
        | TOKEN_mag '(' exp ')'       { $$ = new Foam::volScalarField(Foam::mag(*$3)); delete $3; }
        | TOKEN_magSqrGradGrad '(' exp ')'       { $$ = new Foam::volScalarField(Foam::fvc::magSqrGradGrad(*$3)); delete $3; }
        | TOKEN_mag '(' vexp ')'      { $$ = new Foam::volScalarField(Foam::mag(*$3)); delete $3; }
        | TOKEN_div '(' vexp ')'      { $$ = new Foam::volScalarField(Foam::fvc::div(*$3)); delete $3; }
        | TOKEN_div '(' fsexp ')'      { $$ = new Foam::volScalarField(Foam::fvc::div(*$3)); delete $3; }
        | TOKEN_div '(' fsexp ',' exp ')'      { $$ = new Foam::volScalarField(Foam::fvc::div(*$3,*$5)); delete $3; delete $5; }
        | TOKEN_laplacian '(' exp ')'      { $$ = new Foam::volScalarField(Foam::fvc::laplacian(*$3)); delete $3; }
        | TOKEN_laplacian '(' exp ',' exp ')'      { $$ = new Foam::volScalarField(Foam::fvc::laplacian(*$3,*$5)); delete $3; delete $5; }
        | TOKEN_laplacian '(' fsexp ',' exp ')'      { $$ = new Foam::volScalarField(Foam::fvc::laplacian(*$3,*$5)); delete $3; delete $5; }
        | TOKEN_average '(' fsexp ')'        { $$ = new Foam::volScalarField(Foam::fvc::average(*$3)); delete $3; }
        | TOKEN_integrate '(' fsexp ')'        { $$ = new Foam::volScalarField(Foam::fvc::surfaceIntegrate(*$3)); delete $3; }
        | TOKEN_sum '(' fsexp ')'        { $$ = new Foam::volScalarField(Foam::fvc::surfaceSum(*$3)); delete $3; }
        | vexp '.' 'x'            { $$ = new Foam::volScalarField($1->component(0)); delete $1; }
        | vexp '.' 'y'            { $$ = new Foam::volScalarField($1->component(1)); delete $1; }
        | vexp '.' 'z'            { $$ = new Foam::volScalarField($1->component(2)); delete $1; }
        | lexp '?' exp ':' exp        { $$ = driver.doConditional($1,$3,$5,driver.makeScalarField(0.)); delete $1; delete $3; delete $5; }
        | TOKEN_pi { $$ = driver.makeScalarField(M_PI); }
        | TOKEN_dist '(' ')'        { $$ = driver.makeDistanceField(); }
        | TOKEN_volume '(' ')'        { $$ = driver.makeVolumeField(); }
        | TOKEN_rand '(' ')'        { $$ = driver.makeRandomField(); }
        | TOKEN_randNormal '(' ')'        { $$ = driver.makeGaussRandomField(); }
	| TOKEN_SID		{ 
            $$=driver.getScalarField(*$1);
				}
;

lexp: TOKEN_TRUE   { $$ = driver.makeScalarField(1); }
    | TOKEN_FALSE  { $$ = driver.makeScalarField(0); }
    | TOKEN_set '(' TOKEN_SETID ')'  { $$ = driver.makeCellSetField(*$3); }
    | exp '<' exp  { $$ = driver.doCompare($1,std::less<Foam::scalar>(),$3);  delete $1; delete $3; }
    | exp '>' exp  { $$ = driver.doCompare($1,std::greater<Foam::scalar>(),$3);  delete $1; delete $3; }
    | exp TOKEN_LEQ exp  { $$ = driver.doCompare($1,std::less_equal<Foam::scalar>(),$3);  delete $1; delete $3; }
    | exp TOKEN_GEQ exp  { $$ = driver.doCompare($1,std::greater_equal<Foam::scalar>(),$3);  delete $1; delete $3; }
    | exp TOKEN_EQ exp  { $$ = driver.doCompare($1,std::equal_to<Foam::scalar>(),$3);  delete $1; delete $3; }
    | exp TOKEN_NEQ exp  { $$ = driver.doCompare($1,std::not_equal_to<Foam::scalar>(),$3);  delete $1; delete $3; }
    | '(' lexp ')'		{ $$ = $2; }
    | lexp TOKEN_AND lexp  { $$ = driver.doLogicalOp($1,std::logical_and<Foam::scalar>(),$3);  delete $1; delete $3; }
    | lexp TOKEN_OR lexp   { $$ = driver.doLogicalOp($1,std::logical_or<Foam::scalar>(),$3);  delete $1; delete $3; }
    | '!' lexp %prec TOKEN_NOT { $$ = driver.doLogicalNot($2); delete $2; }
;

vector: TOKEN_VECTOR '(' exp ',' exp ',' exp ')' {     $$ = driver.makeVectorField($3,$5,$7);  delete $3; delete $5; delete $7}
;

%%

void yy::ValueExpressionParser::error (const yy::ValueExpressionParser::location_type& l,const std::string& m)
{
     driver.error (l, m);
}
